/** @file     mapping.h
 *  @author   Adi Singh
 *  @date     March 2019
 */

#ifndef MAPPING_H
#define MAPPING_H

#define PCL_NO_PRECOMPILE

#include <ros/ros.h>
#include <ros/package.h>

#include <geometry_msgs/Vector3Stamped.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>
#include <tf2_ros/transform_listener.h>

#include <visualization_msgs/Marker.h>

#include <pcl/io/pcd_io.h>
#include <pcl_ros/point_cloud.h>

#include <toposens_msgs/TsPoint.h>
#include <toposens_msgs/TsScan.h>


namespace toposens_pointcloud
{
  static const std::string kPointCloudTopic  = "ts_cloud";

  /** Templatized pointcloud containing points of type pcl::PointXYZINormal.*/
  typedef pcl::PointCloud<pcl::PointXYZINormal> XYZINCloud;

/** @brief Demonstrates basic TF and PCL integrations for TsScan data.
 *  @details Subscribes to a topic publishing TsScans and converts
 *  incoming data into PointCloud messages using PCL, optionally
 *  mapped to another frame of reference. Visual behavior of the
 *  pointcloud (scale, color, direction, lifetime, etc) can be
 *  controlled via the PointCloud2 display in Rviz.
 */
class Mapping
{
  public:

    /** Subscribes to a #kScansTopic and prepares a PointCloud structure
     *  for persistent storage.
     *  @param nh Public nodehandle for pub-sub ops on ROS topics.
     *  @param private_nh Private nodehandle for accessing launch parameters.
     */
    Mapping(ros::NodeHandle nh, ros::NodeHandle private_nh);

    /** Destructor frees memory from heap */
    ~Mapping();

  private:
    /** Converts and broadcasts incoming scans as PointCloud messages.
     *  Maintains a local copy of all accumulated scans in their
     *  corresponding PointCloud form.
     *  @param  msg  Pointer to an incoming TsScan message.
     */
    void _convert(const toposens_msgs::TsScan::ConstPtr& msg);

    /** Maps a point to the reference frame defined in #target_frame.
     *  @param transform StampedTransform from original frame to target frame.
     *  @param  p  PointXYZINormal that is transformed to #target_frame.
     *  @param  h  Header data of TsScan the point originates from.
     */
    void _transform(geometry_msgs::TransformStamped transform,
                    pcl::PointXYZINormal& pt,
                    std_msgs::Header h);

    /** Converts a point from a TsPoint message to a standard PCL point.
     *  @param  i  TsPoint to convert.
     *  @returns   PointXYZINormal created from the given input.
     */
    pcl::PointXYZINormal _convertToXYZINormal(toposens_msgs::TsPoint i);

    /** Publishes normals as markers for visualiation in RViz.
     *  @param  tc  The point cloud whose normals are to be published.
     */
    void _publishNormals(XYZINCloud::ConstPtr tc);

    std::string _scans_topic;  /**< Topic name to look for TsScan messages.*/
    std::string _target_frame; /**< Target frame for scan transformations.*/

    ros::Publisher  _cloud_pub; /**< Handler for publishing PCL messages.*/
    ros::Subscriber _scans_sub; /**< Handler for subscribing to TsScans.*/

    ros::Publisher _marker_pub;  /**< Handler for publishing marker messages.*/
    float _lifetime_normals_vis; /**< Lifetime for visualization of normals.*/
    int _marker_id = 0;          /**< Id for publishing unique markers for normals.*/

    tf2_ros::Buffer             _tf2_buffer;    /**< Stores known frame mapping.*/
    tf2_ros::TransformListener *_tf2_listener;  /**< Listener for frame mapping.*/

    boost::thread *_log_thread; /**< Accumulates pointcloud and periodically saves to file. */
};

} // namespace toposens_pointcloud

#endif
